use std::collections::BTreeMap;

use crate::MeanSD;
use crate::StandardDeviationMode;

#[derive(Clone, Copy, Debug, Default, Eq, Ord, PartialEq, PartialOrd)]
/// A bin.
pub struct Bin {
    bottom: u64,
    top: u64,
}

impl Bin {
    /// Returns the bottom end.
    pub fn bottom(&self) -> u64 {
        self.bottom
    }

    /// Returns the top end.
    pub fn top(&self) -> u64 {
        self.top
    }
}

/// An online algorithm for calculating mean and standard deviation with
/// binning and fixed bin sizes.
///
/// The memory requirement is effectively constant, growing only ever so
/// slightly with each new bin.
#[derive(Clone, Debug)]
pub struct Binned {
    all: MeanSD,
    data: BTreeMap<u64, MeanSD>,
    width: u64,
    sd_mode: StandardDeviationMode,
}

impl Binned {
    /// Returns an empty `Binned` structure with `width` bin width.
    ///
    /// **Note:** The bins always use sample standard deviation, `sd_mode` only
    /// applies to `MeanSD` for all observations.
    pub fn new(width: u64, sd_mode: StandardDeviationMode) -> Binned {
        Binned {
            all: MeanSD::new(sd_mode),
            data: BTreeMap::new(),
            width,
            sd_mode,
        }
    }

    /// Returns the bin width.
    pub fn width(&self) -> u64 {
        self.width
    }

    /// Returns a reference to the stats for all observations.
    pub fn all(&self) -> &MeanSD {
        &self.all
    }

    /// Returns an iterator over the bins.
    pub fn bins(&self) -> impl Iterator<Item = (Bin, &MeanSD)> {
        self.data.iter().map(move |(index, meansd)| {
            let bottom = self.width * index;
            let top = self.width * (index + 1) - 1;

            (Bin { bottom, top }, meansd)
        })
    }

    /// Updates the data structure with a new value `x`.
    pub fn update(&mut self, x: f64) {
        self.all.update(x);

        let bin_index = (x / self.width as f64).floor() as u64;

        self.data
            .entry(bin_index)
            .or_insert_with(|| MeanSD::new(StandardDeviationMode::Sample))
            .update(x);
    }
}

// ----------------------------------------------------------------------------
// tests
// ----------------------------------------------------------------------------

#[cfg(test)]
mod tests {
    use super::*;

    use float_cmp::approx_eq;

    #[test]
    fn small() {
        let mut binned = Binned::new(5, StandardDeviationMode::Sample);

        binned.update(2.0);
        binned.update(3.0);
        binned.update(4.0);
        binned.update(5.0);
        binned.update(6.0);
        binned.update(7.0);

        let (n, mean, sd) = binned.data.get(&0).unwrap().tupled();
        assert_eq!(n, 3);
        assert!(approx_eq!(f64, 3.0, mean));
        assert!(approx_eq!(f64, 1.0, sd));

        let (n, mean, sd) = binned.data.get(&1).unwrap().tupled();
        assert_eq!(n, 3);
        assert!(approx_eq!(f64, 6.0, mean));
        assert!(approx_eq!(f64, 1.0, sd));

        let (n, mean, sd) = binned.all.tupled();
        assert_eq!(n, 6);
        assert!(approx_eq!(f64, 4.5, mean));
        assert!(approx_eq!(f64, 1.870_828_693_386_970_7, sd));
    }
}
