use atty::Stream;
use clap::{crate_description, crate_name, crate_version};
use clap::{App, AppSettings, Arg};

use crate::config::ErrorHandler;

pub fn build() -> App<'static, 'static> {
    let color = if atty::is(Stream::Stdout) {
        AppSettings::ColoredHelp
    } else {
        AppSettings::ColorNever
    };

    // ------------------------------------------------------------------------
    // arguments
    // ------------------------------------------------------------------------

    let bin_width = Arg::with_name("bin_width")
        .takes_value(true)
        .short("w")
        .long("bin-width")
        .value_name("width")
        .help("group in bins of width size")
        .long_help(
"Groups output in bins of width size. No empty bins will be shown."
        )
        .validator(positive_integer);

    let handler = Arg::with_name("handler")
        .long("error-handler")
        .hidden_short_help(true)
        .long_help(
"Specifies error handler when input line is not a number. Panic will report \
 the error and quit with failure. Skip will completely ignore the line \
 without printing anything. Warn will ignore just like Skip but show a \
 warning for each failed line.",
        )
        .takes_value(true)
        .case_insensitive(true)
        .possible_values(&ErrorHandler::variants())
        .default_value("Warn");

    let progress = Arg::with_name("progress")
        .takes_value(true)
        .long("progress")
        .value_name("n")
        .hidden_short_help(true)
        .long_help("Print progress every n numbers to STDERR.")
        .validator(positive_integer);

    // ------------------------------------------------------------------------
    // population vs sample
    // ------------------------------------------------------------------------

    let population = Arg::with_name("population")
        .takes_value(false)
        .multiple(true)
        .long("population")
        .help("calculate population standard deviation")
        .overrides_with("sample")
        .display_order(0);

    let sample = Arg::with_name("sample")
        .takes_value(false)
        .multiple(true)
        .long("sample")
        .help("calculate sample standard deviation (default)")
        .overrides_with("population")
        .display_order(0);

    // ------------------------------------------------------------------------
    // put it all together
    // ------------------------------------------------------------------------

    App::new(crate_name!())
        .version(crate_version!())
        .about(crate_description!())
        .global_setting(color)
        .help_short("?")
        .help_message("show this help output")
        .version_message("show version")
        .arg(bin_width)
        .arg(handler)
        .arg(progress)
        .arg(population)
        .arg(sample)
}

#[allow(clippy::needless_pass_by_value)]
fn positive_integer(s: String) -> Result<(), String> {
    if s.parse::<u64>().is_ok() {
        Ok(())
    } else {
        Err(format!("not a positive number: {}", s))
    }
}
