#![deny(clippy::all)]
#![warn(clippy::pedantic, clippy::nursery, clippy::cargo)]

mod cli;
mod config;
mod log;

use atty::Stream;
use prettytable::format::FormatBuilder;
use prettytable::{cell, Row, Table};
use std::io::{self, BufRead};
use std::sync::atomic::{AtomicBool, Ordering};
use std::sync::Arc;

use meansd::{Binned, MeanSD};

use crate::config::Config;
use crate::config::ErrorHandler;

fn main() {
    let args = cli::build().get_matches();
    let config = Config::from_args(&args);

    if atty::is(Stream::Stdin) {
        log::warn("input is read from the terminal");
        log::warn("pipe data into this app");
        log::warn("press CTRL-D to exit");
    };

    let stdin = io::stdin();
    let lines = stdin.lock().lines();

    let numbers = lines.flat_map(|line| line_to_number(line, config));

    if let Some(width) = config.bin_width {
        meansd_binned(numbers, width, config);
    } else {
        meansd(numbers, config);
    };
}

fn line_to_number(line: io::Result<String>, config: Config) -> Option<f64> {
    let line = line.unwrap();
    let line = line.trim();

    match line.parse::<f64>() {
        Ok(v) => Some(v),

        Err(e) => match config.error_handler {
            ErrorHandler::Panic => {
                log::error(format!("{}: {}", e, line));
                std::process::exit(1);
            }

            ErrorHandler::Skip => None,

            ErrorHandler::Warn => {
                log::warn(format!("{}: {}", e, line));
                None
            }
        },
    }
}

fn meansd<I>(numbers: I, config: Config)
where
    I: Iterator<Item = f64>,
{
    let mut meansd = MeanSD::new(config.sd_mode);

    let interrupt = Arc::new(AtomicBool::new(false));
    let interrupt_setter = interrupt.clone();

    ctrlc::set_handler(move || {
        interrupt_setter.store(true, Ordering::SeqCst);
    })
    .expect("error setting interrupt handler");

    for x in numbers {
        meansd.update(x);

        if interrupt.load(Ordering::SeqCst) {
            println!();
            break;
        }

        if config.progress.map_or(false, |p| meansd.size() % p == 0) {
            log::progress(meansd.size(), format!("{}", meansd));
        }
    }

    println!("{}", meansd);
}

pub fn meansd_binned<I>(numbers: I, width: u64, config: Config)
where
    I: Iterator<Item = f64>,
{
    let mut binned = Binned::new(width, config.sd_mode);

    let interrupt = Arc::new(AtomicBool::new(false));
    let interrupt_setter = interrupt.clone();

    ctrlc::set_handler(move || {
        interrupt_setter.store(true, Ordering::SeqCst);
    })
    .expect("error setting interrupt handler");

    for x in numbers {
        binned.update(x);

        if interrupt.load(Ordering::SeqCst) {
            println!();
            break;
        }

        if config
            .progress
            .map_or(false, |p| binned.all().size() % p == 0)
        {
            for (bin, meansd) in binned.bins() {
                let size = binned.all().size();

                log::progress(
                    size,
                    format!(
                        "{} - {} \u{2192} {}",
                        bin.bottom(),
                        bin.top(),
                        meansd
                    ),
                );
            }
        }
    }

    let mut table = Table::new();
    let format = FormatBuilder::new().column_separator(' ').build();
    table.set_format(format);

    let mut titles = Row::empty();

    titles.add_cell(cell!(bu->"From"));
    titles.add_cell(cell!(bur->"To"));
    titles.add_cell(cell!(bur->"Size"));
    titles.add_cell(cell!(bur->"Mean"));
    titles.add_cell(cell!(bur->"SD"));

    table.set_titles(titles);

    for (bin, meansd) in binned.bins() {
        let size = meansd.size();
        let mean = meansd.mean().round();
        let sd = meansd.sd().round();

        let mut row = Row::empty();
        row.add_cell(cell!(r->bin.bottom()));
        row.add_cell(cell!(r->bin.top()));
        row.add_cell(cell!(r->size));
        row.add_cell(cell!(r->mean));
        row.add_cell(cell!(r->sd));

        table.add_row(row);
    }

    println!();
    table.printstd();
    println!();
    println!("{}", binned.all());
}
