use std::fmt;

use crate::StandardDeviationMode;

/// An online algorithm for calculating mean and standard deviation.
///
/// The memory requirement is constant.
#[derive(Clone, Copy, Debug, Default)]
pub struct MeanSD {
    sd_mode: StandardDeviationMode,
    size: u64,
    mean: f64,
    m2: f64,
}

impl MeanSD {
    /// Returns an empty `MeanSD` structure.
    pub fn new(sd_mode: StandardDeviationMode) -> MeanSD {
        let mut meansd = MeanSD::default();
        meansd.sd_mode = sd_mode;
        meansd
    }

    /// Returns the number of observations.
    pub fn size(&self) -> u64 {
        self.size
    }

    /// Returns the mean.
    pub fn mean(&self) -> f64 {
        self.mean
    }

    /// Returns the standard deviation.
    pub fn sd(&self) -> f64 {
        if self.size > 1 {
            let div = match self.sd_mode {
                StandardDeviationMode::Population => self.size as f64,
                StandardDeviationMode::Sample => self.size as f64 - 1.0,
            };

            (self.m2 / div).sqrt()
        } else {
            0.0
        }
    }

    /// Returns the standard error of the mean.
    pub fn sem(&self) -> f64 {
        self.sd() / (self.size as f64).sqrt()
    }

    /// Updates the data structure with a new value `x`.
    pub fn update(&mut self, x: f64) {
        let delta = x - self.mean;

        self.size += 1;
        self.mean += delta / (self.size as f64);
        self.m2 += delta * (x - self.mean);
    }

    #[cfg(test)]
    pub(crate) fn tupled(&self) -> (u64, f64, f64) {
        (self.size, self.mean, self.sd())
    }
}

impl fmt::Display for MeanSD {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let n = self.size();
        let mean = self.mean().round();
        let sd = self.sd().round();

        write!(f, "n={} ∅ {} ± {}", n, mean, sd)
    }
}

// ----------------------------------------------------------------------------
// tests
// ----------------------------------------------------------------------------

#[cfg(test)]
mod tests {
    use super::*;

    use float_cmp::approx_eq;

    #[test]
    fn empty() {
        let meansd = MeanSD::default();
        let (n, mean, sd) = meansd.tupled();

        assert_eq!(n, 0);
        assert!(approx_eq!(f64, 0.0, mean));
        assert!(approx_eq!(f64, 0.0, sd));
    }

    #[test]
    fn single() {
        let mut meansd = MeanSD::default();

        meansd.update(1.0);

        let (n, mean, sd) = meansd.tupled();

        assert_eq!(n, 1);
        assert!(approx_eq!(f64, 1.0, mean));
        assert!(approx_eq!(f64, 0.0, sd));
    }

    #[test]
    fn small() {
        let mut meansd = MeanSD::default();

        meansd.update(1.0);
        meansd.update(2.0);
        meansd.update(3.0);

        let (n, mean, sd) = meansd.tupled();

        assert_eq!(n, 3);
        assert!(approx_eq!(f64, 2.0, mean));
        assert!(approx_eq!(f64, 1.0, sd));
    }

    #[test]
    fn population() {
        let mut meansd = MeanSD::new(StandardDeviationMode::Population);

        meansd.update(2.0);
        meansd.update(4.0);
        meansd.update(4.0);
        meansd.update(4.0);
        meansd.update(5.0);
        meansd.update(5.0);
        meansd.update(7.0);
        meansd.update(9.0);

        let (n, mean, sd) = meansd.tupled();

        assert_eq!(n, 8);
        assert!(approx_eq!(f64, 5.0, mean));
        assert!(approx_eq!(f64, 2.0, sd));
    }
}
